@@ -6,31 +6,9 @@ from collections import namedtuple
 import numpy as np
 from graphdot.graph import Graph
 from graphdot.util import Timer
+from graphdot.util.iterable import fold_like, flatten
 from ._backend_factory import backend_factory
 
-__all__ = ['MarginalizedGraphKernel']
-
-
-def flatten(iterable):
-    for item in iterable:
-        if hasattr(item, '__iter__'):
-            yield from flatten(item)
-        else:
-            yield item
-
-
-def fold_like(flat, example):
-    folded = []
-    for item in example:
-        if hasattr(item, '__iter__'):
-            n = len(list(flatten(item)))
-            folded.append(fold_like(flat[:n], item))
-            flat = flat[n:]
-        else:
-            folded.append(flat[0])
-            flat = flat[1:]
-    return tuple(folded)
-
 
 class MarginalizedGraphKernel:
     """Implements the random walk-based graph similarity kernel as proposed in:
@@ -40,9 +18,9 @@ class MarginalizedGraphKernel:
 
     Parameters
     ----------
-    node_kernel: base kernel or composition of base kernels
+    node_kernel: microkernel
         A kernelet that computes the similarity between individual nodes
-    edge_kernel: base kernel or composition of base kernels
+    edge_kernel: microkernel
         A kernelet that computes the similarity between individual edge
     kwargs: optional arguments
         p: functor or 'uniform' or 'default'
@@ -102,6 +80,9 @@ class MarginalizedGraphKernel:
             The graphs must all have same node and edge attributes.
         Y: None or list of M graphs
             The graphs must all have same node and edge attributes.
+        eval_gradient: Boolean
+            If True, computes the gradient of the kernel matrix with respect
+            to hyperparameters and return it alongside the kernel matrix.
         nodal: bool
             If True, return node-wise similarities; otherwise, return graphwise
             similarities.
@@ -115,10 +96,13 @@ class MarginalizedGraphKernel:
 
         Returns
         -------
-        numpy.array
+        kernel_matrix: ndarray
             if Y is None, return a square matrix containing pairwise
             similarities between the graphs in X; otherwise, returns a matrix
             containing similarities across graphs in X and Y.
+        gradient: ndarray
+            The gradient of the kernel matrix with respect to kernel
+            hyperparameters. Only returned if eval_gradient is True.
         """
         timer = Timer()
         backend = self.backend
@@ -223,13 +207,16 @@ class MarginalizedGraphKernel:
         else:
             return output.astype(self.element_dtype)
 
-    def diag(self, X, nodal=False, lmin=0, timing=False):
+    def diag(self, X, eval_gradient=False, nodal=False, lmin=0, timing=False):
         """Compute the self-similarities for a list of graphs
 
         Parameters
         ----------
         X: list of N graphs
             The graphs must all have same node attributes and edge attributes.
+        eval_gradient: Boolean
+            If True, computes the gradient of the kernel matrix with respect
+            to hyperparameters and return it alongside the kernel matrix.
         nodal: bool
             If True, returns a vector containing nodal self similarties; if
             False, returns a vector containing graphs' overall self
@@ -251,9 +238,18 @@ class MarginalizedGraphKernel:
             if nodal=False, returns a vector containing graphs' overall self
             similarities; if nodal = 'block', return a list of square matrices,
             each being a pairwise nodal similarity matrix within a graph.
+        gradient
+            The gradient of the kernel matrix with respect to kernel
+            hyperparameters. Only returned if eval_gradient is True.
         """
         timer = Timer()
         backend = self.backend
+        traits = self.traits(
+            diagonal=True,
+            nodal=nodal,
+            lmin=lmin,
+            eval_gradient=eval_gradient
+        )
 
         ''' assert graph attributes are compatible with each other '''
         pred_or_tuple = Graph.has_unified_types(X)
@@ -294,7 +290,11 @@ class MarginalizedGraphKernel:
             output_length = int(starts[-1])
         else:
             raise(ValueError("Invalid 'nodal' option '%s'" % nodal))
-        output = backend.empty(output_length, np.float32)
+        if traits.eval_gradient is True:
+            output_shape = (output_length, 1 + self.n_dims)
+        else:
+            output_shape = (output_length, 1)
+        output = backend.empty(int(np.prod(output_shape)), np.float32)
         timer.toc('creating output buffer')
 
         ''' call GPU kernel '''
@@ -308,8 +308,8 @@ class MarginalizedGraphKernel:
             jobs,
             starts,
             output,
-            (output_length, 1),
-            self.traits(diagonal=True, nodal=nodal, lmin=lmin),
+            output_shape,
+            traits,
             timer,
         )
         timer.toc('calling GPU kernel (overall)')
@@ -325,7 +325,13 @@ class MarginalizedGraphKernel:
             timer.report(unit='ms')
         timer.reset()
 
-        return output.astype(self.element_dtype)
+        if traits.eval_gradient is True:
+            return (
+                output[:, 0].astype(self.element_dtype),
+                output[:, 1:].astype(self.element_dtype)
+            )
+        else:
+            return output.astype(self.element_dtype)
 
     """⭣⭣⭣⭣⭣ scikit-learn interoperability methods ⭣⭣⭣⭣⭣"""
 
